/* ELF Dump
 * Utility to analyse data structures of ELF files.
 * Written by Owen Klan  -  August 2002 */

/* TODO:
 */

#include <stdio.h>
#include <elf.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>		       /* for open() */
#include <sys/mman.h>		       /* for mmap() */
#include <sys/stat.h>		       /* for stat() */
#include <errno.h>
#include <unistd.h>

#include "elfdump.h"

#define MAX_FILEPATH    128

/* function prototypes */
static int valid_elf(char *image);
static char *get_arch(Elf32_Ehdr *header);
static char *get_type(Elf32_Ehdr *header);
char *get_seg_type(Elf32_Phdr *header);
static char *get_interpreter(Elf32_Ehdr *header, char *image);
static void display_header(Elf32_Ehdr *header, char *elf_path);
static void load_names(Elf32_Ehdr *header, char *image);
static void usage();

/* global sections name table */
char *name_table = NULL;
/* global name table for the symbol table */
char *sym_name_table = NULL;

unsigned char byte_order;

int main(int argc, char *argv[])
{
    Elf32_Ehdr *elf_head;	       /* ELF header structure */
    char elf_path[MAX_FILEPATH];       /* path for the file */
    int elf_fp; 		       /* fd for the file */
    struct stat elf_stat;	       /* used with stat() */
    char *elf_image;		       /* used with mmap() */
    
    if (argc < 3) {
	usage();
    }
    
    /* 
     * get the filename and attempt to open it
     */
    strncpy(elf_path, argv[1], MAX_FILEPATH);
    
    if ((elf_fp = open(elf_path, O_RDONLY)) < 0) {
	fprintf(stderr, "Failed opening %s! %s\n", elf_path,
		strerror(errno));
	exit(1);
    }
    
    /*
     * use stat() to get the file size
     */
    if (stat(elf_path, &elf_stat) < 0) {
	fprintf(stderr, "Failed determining file size! %s\n",
		strerror(errno));
	exit(1);
    }
    
    /*
     * attempt to mmap() the file to an address
     */
    if ((elf_image = mmap(0, elf_stat.st_size, PROT_READ, MAP_PRIVATE,
			  elf_fp, 0)) == MAP_FAILED) {
	fprintf(stderr, "Failed mapping ELF file to memory! %s\n",
		strerror(errno));
	exit(1);
    }
    
    /* 
     * check it's a valid ELF file, close and exit if not
     */
    if (!valid_elf(elf_image)) {
	fprintf(stderr, "%s is not a valid ELF file!\n", elf_path);
	exit(1);
    }
    
    /* 
     * fill the header structure
     */
    elf_head = (Elf32_Ehdr *)elf_image;
    byte_order = elf_head->e_ident[EI_DATA];
    
    display_header(elf_head, elf_path);
    
    /*
     * load the string table pointed out by the segment header
     */
    load_names(elf_head, elf_image);
    
    /* determine what data to process */
    {
	int cur_arg = 2;
	
	for (;cur_arg < argc;cur_arg++) {
	    if (strncmp(argv[cur_arg], "--sections", 10) == 0) {
		display_sections(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--secrep", 8) == 0) {
		sections_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--segrep", 8) == 0) {
		segments_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--symrep", 8) == 0) {
		symbols_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--segments", 10) == 0) {
		display_segments(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--dynamic", 9) == 0) {
		dynamic_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--dynsym", 8) == 0) {
		dynsym_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--interp", 8) == 0) {
		fprintf(stdout, "\nProgram Interpreter:\t%s\n",
			get_interpreter(elf_head, elf_image));
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--depend", 8) == 0) {
		display_depends(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--symbols", 11) == 0) {
		display_symbols(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--funcrep", 9) == 0) {
		functions_report(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--files", 7) == 0) {
		display_files(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--got", 5) == 0) {
		display_got(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--exports", 9) == 0) {
		display_exports(elf_head, elf_image);
		continue;
	    }
	    if (strncmp(argv[cur_arg], "--reloc", 7) == 0) {
		reloc_report(elf_head, elf_image);
		continue;
	    }
	}
    }
    
    if (munmap(elf_image, elf_stat.st_size) < 0) {
	fprintf(stderr, "Failed unmapping ELF file!\n");
	fprintf(stderr, "%s\n", strerror(errno));
	exit(1);
    }
    
    return 0;
}


/* Function to swap the byte-order for Elf32_Half items (ushorts) */
Elf32_Half ehalf_swap(Elf32_Half half)
{
    unsigned char ha[2];
#ifdef BIG_ENDIAN
    if (byte_order != ELFDATA2MSB)
      return half;
#else
    if (byte_order != ELFDATA2LSB)
      return half;
#endif
    
    ha[0] = (half & 0xff00) >> 8;
    ha[1] = (half & 0x00ff);
    return *(Elf32_Half *)ha;
}

/* Function to swap the byte-order for Elf32_Word items (ints) */
Elf32_Word eword_swap(Elf32_Word word)
{
    unsigned char wa[4];
    
#ifdef BIG_ENDIAN
    if (byte_order != ELFDATA2MSB)
      return word;
#else
    if (byte_order != ELFDATA2LSB)
      return word;
#endif
    
    wa[0] = (word & 0xff000000) >> 24;
    wa[1] = (word & 0x00ff0000) >> 16;
    wa[2] = (word & 0x0000ff00) >> 8;
    wa[3] = (word & 0x000000ff);
    
    return *(Elf32_Word *)wa;
}


/* Procedure to display usage information */
static void usage()
{
    fprintf(stderr, "Usage:     elfdump <file> <output data>\n");
    fprintf(stderr, "\nValid options for <output data> are:\n");
    fprintf(stderr, "  --basic       Displays just the ELF version, file "
	    "type and target architecture\n");
    fprintf(stderr, "  --depend      Displays Shared Object dependency"
	    " information\n");
    fprintf(stderr, "  --dynamic     Displays information regarding dynamic"
	    " linking requirements\n");
    fprintf(stderr, "  --dynsym      Displays information regarding dynamic"
	    " linking symbols\n");
    fprintf(stderr, "  --exports     Displays exported symbol information\n");
    fprintf(stderr, "  --files       Displays file information for the"
	    " binary\n");
    fprintf(stderr, "  --funcrep     Displays information about functions"
	    " in a column format\n");
    fprintf(stderr, "  --got         Displays the values contained within the"
	    " Global Offset Table\n");
    fprintf(stderr, "  --interp      Displays the name of the executable"
	    " that interprets this file\n");
    fprintf(stderr, "  --reloc       Displays relocation information"
	    " (no addends)\n");
    fprintf(stderr, "  --secrep      Displays section information in a"
	    " column format\n");
    fprintf(stderr, "  --sections    Displays section information\n");
    fprintf(stderr, "  --segments    Displays segment information\n");
    fprintf(stderr, "  --segrep      Displays segment information in a"
	    " column format\n");
    fprintf(stderr, "  --symbols     Displays symbol information\n");
    fprintf(stderr, "  --symrep      Displays symbol information in a"
	    " column format\n");
    fprintf(stderr, "\nNote that the following options only work for"
	    " non-stripped binaries:\n");
    fprintf(stderr, "  --symbols, --symrep, --exports, --funcrep, --files\n");
    exit(1);
}


/* Displays the basic ELF header information such as machine
 * architecture, ELF version and such */
static void display_header(Elf32_Ehdr *header, char *elf_path)
{
    char data_encoding[16];
    
    if (header->e_ident[EI_DATA] == ELFDATA2LSB)
      strncpy(data_encoding, "Little Endian", 16);
    else if (header->e_ident[EI_DATA] == ELFDATA2MSB)
      strncpy(data_encoding, "Big Endian", 16);
    
    fprintf(stdout, "ELFdump "VERSION"  -  Written by Owen Klan 2002\n");
    fprintf(stdout, "ELF Statistics for %s, %s ELF version %d file\n", 
	    elf_path, data_encoding, eword_swap(header->e_version));
    fprintf(stdout, "File type:     %s\n", get_type(header));
    fprintf(stdout, "Architecture:  %s\n", get_arch(header));
    fprintf(stdout, "Entry point:   0x%08X\tCPU-specific flags:   0x%08X\n", 
	    eword_swap(header->e_entry), eword_swap(header->e_flags));
    fprintf(stdout, "Segment Table:\n");
    fprintf(stdout, "File Offset:  0x%08X  Entry Size:  %d  Num. Entries:  %d"
	    "  Table Size:  %d\n",
	    eword_swap(header->e_phoff), ehalf_swap(header->e_phentsize),
	    ehalf_swap(header->e_phnum),
	    (ehalf_swap(header->e_phentsize) * ehalf_swap(header->e_phnum)));
    fprintf(stdout, "Section Table:\n");
    fprintf(stdout, "File Offset:  0x%08X  Entry Size:  %d  Num. Entries:  %d"
	    "  Table Size:  %d\n",
	    eword_swap(header->e_shoff), ehalf_swap(header->e_shentsize),
	    ehalf_swap(header->e_shnum),
	    (ehalf_swap(header->e_shentsize) * ehalf_swap(header->e_shnum)));
}

/* Loads the string table section pointed to by the ELF header's
 * e_shstrndx field and also loads the symbol table's string table */
static void load_names(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *ptr;
    int i;
    
    /* load the section names string table */
    ptr = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    
    for (i=0;i < ehalf_swap(header->e_shstrndx); i++)
      ptr++;
    name_table = (char *)(image + eword_swap(ptr->sh_offset));

    /* load the general string table */
    ptr = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    for (i=0; i < (int)ehalf_swap(header->e_shnum); i++) {
	if (strncmp((name_table + eword_swap(ptr->sh_name)),
		    ".strtab", 7) == 0) {
	    sym_name_table = (char *)(image + eword_swap(ptr->sh_offset));
	    break;
	} else {
	    ptr++;
	}
    }
    
    if (sym_name_table == NULL) {
	fprintf(stdout, "\nNo symbol table defined!\n\n");
    }
}


/* Reads the 'Program Interpreter' section from the ELF file and
 * return's it as a string. NOTE: DO NOT FREE THE RETURNED STRING! */
static char *get_interpreter(Elf32_Ehdr *header, char *image)
{
    Elf32_Phdr *segtab, *ptr;
    int segtab_size;
    int segtab_offset;
    int num_segs;
    int i;
    
    num_segs = ehalf_swap(header->e_phnum);
    segtab_size = ehalf_swap(header->e_phentsize) * ehalf_swap(header->e_phnum);
    segtab_offset = eword_swap(header->e_phoff);
    
    /* 
     * load the address of the segment table
     */
    segtab = (Elf32_Phdr *)(image + segtab_offset);
    ptr = segtab;
   
    /* find the program interpreter segment from the segments table
     * and attempt to read the segment as a string */
    for (i=0;i<num_segs;i++) {
	if (strncmp(get_seg_type(ptr), "Program Interpreter", 19) == 0) {
	    /* we have the interpreter segment, woohoo! */
	    int interp_offset;
	    int interp_size;
	    
	    interp_offset = eword_swap(ptr->p_offset);
	    interp_size = eword_swap(ptr->p_filesz);
	    
	    return (char *)(image + interp_offset);
	}
	ptr++;
    }
    
    return NULL;
}


/* Determine the file's type. ie. relocatable, executable etc, etc. */
static char *get_type(Elf32_Ehdr *header)
{
    static char type_text[32];
    
    memset(type_text, 0x00, 32);
    
    /* first check to see if the type value lies withing the
     * OS-specific or CPU-specific regions */
    if (ehalf_swap(header->e_type) >= 0xFE00 
     && ehalf_swap(header->e_type) <= 0xFEFF) {
	sprintf(type_text, "OS specific: %d", ehalf_swap(header->e_type));
	return type_text;
    }
    if (ehalf_swap(header->e_type) >= 0xFF00
     && ehalf_swap(header->e_type) <= 0xFFFF) {
	sprintf(type_text, "CPU specific: %d", ehalf_swap(header->e_type));
	return type_text;
    }
    
    switch (ehalf_swap(header->e_type)) {
      case ET_NONE:
	strncpy(type_text, "No type?", 32); break;
      case ET_REL:
	strncpy(type_text, "Relocatable file", 32); break;
      case ET_EXEC:
	strncpy(type_text, "Executable file", 32); break;
      case ET_DYN:
	strncpy(type_text, "Shared object file", 32); break;
      case ET_CORE:
	strncpy(type_text, "Core file",32); break;
      default:
	sprintf(type_text, "Unknown: %d", ehalf_swap(header->e_type));
    };
    
    return type_text;
}

/* Determine's the target architecture and returns an ASCII
 * representation of it. Currently this function only translates
 * the more common machine architectures. A return of "Unknown"
 * doesn't necessarily mean the architecture is invalid, get_arch() 
 * just doesn't know about it :) */
static char *get_arch(Elf32_Ehdr *header)
{
    static char arch_text[32];
    
    memset(arch_text, 0x00, 32);
    
    switch (ehalf_swap(header->e_machine)) {
    case EM_NONE:
	strncpy(arch_text, "No Machine", 32); break;
#ifdef EM_M32	
    case EM_M32:
	strncpy(arch_text, "AT&T WE 32100", 32); break;
#endif
#ifdef EM_SPARC
    case EM_SPARC:
	strncpy(arch_text, "Sun SPARC", 32); break;
#endif
#ifdef EM_386
    case EM_386:
	strncpy(arch_text, "Intel 80386", 32); break;
#endif
#ifdef EM_68K
    case EM_68K:
	strncpy(arch_text, "Motorola M68K", 32); break;
#endif
#ifdef EM_88K
    case EM_88K:
	strncpy(arch_text, "Motorola M88K", 32); break;
#endif
#ifdef EM_860
    case EM_860:
	strncpy(arch_text, "Intel 80860", 32); break;
#endif
#ifdef EM_MIPS
    case EM_MIPS:
	strncpy(arch_text, "MIPS R3000 Big-endian", 32); break;
#endif
#ifdef EM_MIPS_RS3_LE
    case EM_MIPS_RS3_LE:
	strncpy(arch_text, "MIPS R3000 Little-endian", 32); break;
#endif
#ifdef EM_S370
    case EM_S370:
	strncpy(arch_text, "IBM System/370", 32); break;
#endif
#ifdef EM_960	
    case EM_960:
	strncpy(arch_text, "Intel 80960", 32); break;
#endif
#ifdef EM_PPC
    case EM_PPC:
	strncpy(arch_text, "PowerPC", 32); break;
#endif
#ifdef EM_PPC64
    case EM_PPC64:
	strncpy(arch_text, "PowerPC 64-bit", 32); break;
#endif
#ifdef EM_S390
    case EM_S390:
	strncpy(arch_text, "IBM S390", 32); break;
#endif
#ifdef EM_V800
    case EM_V800:
	strncpy(arch_text, "NEC V800", 32); break;
#endif
#ifdef EM_ARM
    case EM_ARM:
	strncpy(arch_text, "ARM", 32); break;
#endif
#ifdef EM_FAKE_ALPHA
    case EM_FAKE_ALPHA:
	strncpy(arch_text, "DEC Alpha", 32); break;
#endif
    default:
	sprintf(arch_text, "Unknown: %d", ehalf_swap(header->e_machine));
    }
    
    return arch_text;
}


/* Validates that the file is actually an ELF file
 * by checking the first four bytes. Returns zero if it isn't
 * an ELF file, non-zero if it is */
static int valid_elf(char *image)
{
    unsigned char *magic;   /* hold the magic number */

    magic = (unsigned char *)image;    /* point to beginning of ELF image */
    
    if ((magic[0] == 0x7F) && (strncmp((magic+1), "ELF", 3) == 0))
      return 1;
    else
      return 0;
}
