/* segments.c - contains the program segment processing
 * functions for elfdump.
 * Written by Owen Klan  -  August 2002 */

#include <stdio.h>
#include <elf.h>
#include <stdlib.h>
#include <string.h>

#include "elfdump.h"

/* Function prototypes */
char *get_seg_type(Elf32_Phdr *header);
char *get_seg_flags(Elf32_Phdr *header);
void dump_segment(Elf32_Phdr *header, int n);


/* Reads and displays the contents of the segments table */
void display_segments(Elf32_Ehdr *header, char *image)
{
    Elf32_Off segtab_offset;	       /* segment table offset */
    Elf32_Half segtab_size;	       /* segment table size */
    Elf32_Half num_segs;	       /* number of segment entries */
    int i = 0, num_unused = 0;
    Elf32_Phdr *seg_table, *ptr;       /* segment table buffer */
    
    segtab_offset = eword_swap(header->e_phoff);
    num_segs = ehalf_swap(header->e_phnum);
    segtab_size = ehalf_swap(header->e_phentsize) * num_segs;
    
    fprintf(stdout, "\nProgram Segment Details");
    
    seg_table = (Elf32_Phdr *)(image + segtab_offset);
    ptr = (Elf32_Phdr *)seg_table;
    
    /* process segments, one at a time */
    for (i=0;i<num_segs;i++) {
	if (eword_swap(ptr->p_type) != PT_NULL)
	  dump_segment(ptr, i);
	else
	  num_unused++;
	
	ptr++;
    }
    if (num_unused > 0) {
	fprintf(stdout, "%d unused segments were not displayed\n",
		num_unused);
    }
}

/* Reads and displays the contents of the segments table in a column
 * format suitable for quick comparison */
void segments_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Off segtab_offset;	       /* segment table offset */
    Elf32_Half segtab_size;	       /* segment table size */
    Elf32_Half num_segs;	       /* number of segment entries */
    int i = 0;
    Elf32_Phdr *seg_table, *ptr;       /* segment table buffer */
    char *flags_str = NULL;
    
    if (!header || !image)
      return;
    
    segtab_offset = eword_swap(header->e_phoff);
    segtab_size = eword_swap(header->e_phentsize) * eword_swap(header->e_phnum);
    num_segs = ehalf_swap(header->e_phnum);
    
    fprintf(stdout, "\nProgram Segment Details\n");
    fprintf(stdout, "Segment  PhysAddr  VirtAddr  FileOff.  %8s  %8s  Flags"
	    "  Alignment\n", "FileSize", "Mem.Size");

//    fprintf(stdout, "Segment PhysAddr  VirtAddr  FileOff.  Size    "
//	    "Flags  Alignment\n");
    
    seg_table = (Elf32_Phdr *)(image + segtab_offset);
    ptr = (Elf32_Phdr *)seg_table;
    
    /* process segments, one at a time */
    for (i=0;i<num_segs;i++) {
	flags_str = get_seg_flags(ptr);
	
	fprintf(stdout, "   %d     %08X  %08X  %08X  %8X  %8X   %s   %d\n",
		i, eword_swap(ptr->p_paddr), eword_swap(ptr->p_vaddr),
		eword_swap(ptr->p_offset),
		eword_swap(ptr->p_filesz),
		eword_swap(ptr->p_memsz),
		flags_str, eword_swap(ptr->p_align));
	ptr++;
        if (flags_str)
	  free(flags_str);
    }
}


/* Procedure to print information about the passed segment to the
 * standard output. */
void dump_segment(Elf32_Phdr *header, int n)
{
    char *type_str = NULL;
    char *flags_str = NULL;
    
    if (header == NULL)
      return;			       /* don't try to dereference a NULL */
    
    type_str = get_seg_type(header);
    flags_str = get_seg_flags(header);
    
    fprintf(stdout, "\nSegment %d   Alignment: %d\t Type:  %s\n", 
	    n, eword_swap(header->p_align), type_str);
    fprintf(stdout, " Physical Address: 0x%08X    Virtual Address: 0x%08X\n",
	    eword_swap(header->p_paddr), eword_swap(header->p_vaddr));
    fprintf(stdout, " Size in File: %d\t\t Size in Memory: %d\n",
	    eword_swap(header->p_filesz), eword_swap(header->p_memsz));
    fprintf(stdout, " Flags:  %s\n", flags_str);
    
    if (type_str)
      free(type_str);
    if (flags_str)
      free(flags_str);
}

/* Returns the ASCII representation of the segment type */
char *get_seg_type(Elf32_Phdr *header)
{
    char *seg_type = NULL;

    if ((seg_type = malloc(32)) == NULL)
      return NULL;
    
    memset(seg_type, 0x00, 32);
    
    switch (eword_swap(header->p_type)) {
      case PT_NULL:
	sprintf(seg_type, "Unused"); break;
      case PT_LOAD:
	sprintf(seg_type, "Loadable Segment"); break;
      case PT_DYNAMIC:
	sprintf(seg_type, "Dynamic Linking Info."); break;
      case PT_INTERP:
	sprintf(seg_type, "Program Interpreter"); break;
      case PT_NOTE:
	sprintf(seg_type, "Auxiliary Information"); break;
      case PT_SHLIB:
	sprintf(seg_type, "Reserved"); break;
      case PT_PHDR:
	sprintf(seg_type, "Segment Table"); break;
      default:
	sprintf(seg_type, "Unknown: %d", eword_swap(header->p_type));
    };
    
    return seg_type;
}

/* Function that will make an ASCII representation of
 * a segment's flags */
char *get_seg_flags(Elf32_Phdr *header)
{
    char *flags_str = NULL;
    Elf32_Word pflags;
    
    if (!header)
      return NULL;
    
    if ((flags_str = malloc(4)) == NULL)
      return NULL;
    sprintf(flags_str, "---");
    *(flags_str + 3) = '\0';
    
    pflags = eword_swap(header->p_flags);
    
    /* parse the flags */
    if ((pflags & PF_X) == PF_X)
      *(flags_str + 2) = 'X';
    if ((pflags & PF_W) == PF_W)
      *(flags_str + 1) = 'W';
    if ((pflags & PF_R) == PF_R)
      *(flags_str) = 'R';
    
    return flags_str;
}
